/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;
import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our sixth rendering pipeline. It adds a view
   volume normalization stage, {@link View2Camera}, that converts
   a {@link Camera}'s configurable view volume into the normalized
   view volume used by the {@link Clip} pipeline stage. Coordinates
   relative to the {@link Camera}'s arbitrary view volume are called
   "view coordinates" and coordinates relative to the normalized view
   volume are called "camera coordinates", so the new pipeline stage
   converts vertex coordinates from view coordinates to camera
   coordinates. There are four pipeline stages.
*/
public final class Pipeline
{
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      // Render every Model in the Scene.
      for (final Model model : scene.modelList)
      {
         if ( model.visible )
         {
            logMessage(model, "==== Render Model: " + model.name + " ====");

            check(model);

            logVertexList("0. View   ", model);

            // 1. Apply the Camera's normalizing view-to-camera coordinate transformation.
            final Model model1 = View2Camera.view2camera(model, scene.camera);

            logVertexList("1. Camera   ", model1);

            // 2. Apply the projection transformation.
            final Model model2 = Projection.project(model1, scene.camera);

            logVertexList("2. Projected", model2);

            // 3. Clip line segments to the camera's view rectangle.
            final Model model3 = Clip.clip(model2);

            logVertexList("3. Clipped  ", model3);
            logColorList("3. Clipped  ", model3);
            logLineSegmentList("3. Clipped  ", model3);

            // 4. Rasterize each visible line segment into pixels.
            for (final LineSegment ls : model3.lineSegmentList)
            {
               logLineSegment("4. Rasterize", model3, ls);

               RasterizeAntialias.rasterize(model3, ls, vp);
            }

            logMessage(model, "==== End Model: " + model.name + " ====");
         }
         else
         {
            logMessage(model, "==== Hidden model: " + model.name + " ====");
         }
      }
   }


   /**
      Determine if there are any obvious problems with the {@link Model}
      being rendered. The purpose of these checks is to make the renderer
      a bit more user friendly. If a user makes a simple mistake and tries
      to render a {@link Model} that is missing vertices, line segments,
      or colors, then the user gets a helpful error message.

      @param model  the {@link Model} being rendered
   */
   public static void check(final Model model)
   {
      boolean error = false;
      if (model.vertexList.isEmpty())
      {
         System.err.println("***WARNING: This model does not have any vertices.");
         error = true;
      }
      if (model.lineSegmentList.isEmpty())
      {
         System.err.println("***WARNING: This model does not have any line segments.");
         error = true;
      }
      if (model.colorList.isEmpty())
      {
         System.err.println("***WARNING: This model does not have any colors.");
         error = true;
      }
      if (error)
      {
         System.err.println(model);
         Thread.dumpStack();
       //System.err.println(Arrays.toString(Thread.currentThread().getStackTrace()));
         System.exit(-1);
      }
   }


   /**
      Use the pipeline's and the model's debug variables to determine
      if the given message should be printeed to stderr.

      @param model    the {@link Model} being rendered
      @param message  {@link String} to output to stderr
   */
   public static void logMessage(final Model model, final String message)
   {
      if (Pipeline.debug || model.debug)
         System.err.println( message );
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Model}'s {@link Vertex} list.

      @param stage  name for the pipeline stage
      @param model  the {@link Model} whose {@link Vertex} list is to be printed
   */
   private static void logVertexList(final String stage, final Model model)
   {
      if (Pipeline.debug || model.debug)
      {
         int i = 0;
         for (final Vertex v : model.vertexList)
         {
            System.err.printf("%s: vIndex = %3d, %s", stage, i, v.toString());
            ++i;
         }
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Model}'s {@link Color} list.

      @param stage  name for the pipeline stage
      @param model  the {@link Model} whose {@link Color} list is to be printed
   */
   private static void logColorList(final String stage, final Model model)
   {
      if (Pipeline.debug || model.debug)
      {
         int i = 0;
         for (final Color c : model.colorList)
         {
            System.err.printf("%s: cIndex = %3d, %s\n", stage, i, c.toString());
            ++i;
         }
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Model}'s {@link LineSegment} list.

      @param stage  name for the pipeline stage
      @param model  the {@link Model} whose line segment list is to be printed
   */
   private static void logLineSegmentList(final String stage, final Model model)
   {
      if (Pipeline.debug || model.debug)
      {
         for (final LineSegment ls : model.lineSegmentList)
         {
            System.err.printf("%s: %s", stage, ls.toString());
         }
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   public static void logLineSegment(final String stage, final Model model, final LineSegment ls)
   {
      if (Pipeline.debug || model.debug)
      {
         System.err.print( stage + ": " + ls.toString() );
         final int index0 = ls.vIndex[0];
         final int index1 = ls.vIndex[1];
         final Vertex v0 = model.vertexList.get(index0);
         final Vertex v1 = model.vertexList.get(index1);
         System.err.printf("   vIndex = %3d, %s", index0, v0.toString());
         System.err.printf("   vIndex = %3d, %s", index1, v1.toString());

         final int cIndex0 = ls.cIndex[0];
         final int cIndex1 = ls.cIndex[1];
         final Color c0 = model.colorList.get(cIndex0);
         final Color c1 = model.colorList.get(cIndex1);
         System.err.printf("   cIndex = %3d, %s\n", cIndex0, c0.toString());
         System.err.printf("   cIndex = %3d, %s\n", cIndex1, c1.toString());
      }
   }
}
