/*

*/

package renderer.pipeline;
import  renderer.scene.*;

import java.util.List;
import java.util.ArrayList;

/**
   Transform each {@link Vertex} of a {@link Model} from
   view coordinates to normalized camera coordinates.
<p>
   This stage transforms the {@link Camera}'s view volume from a
   user defined shape (in the view coordinate system) into the
   standard normalized view volume (in the camera coordinate system)
   used by the {@link Clip} pipeline stage.
<p>
   There are two standard normalized view volumes, one for perspective
   projection and one for orthographic projection.
<p>
   The standard normalized perspective view volume is the infinitely
   long pyramid with its apex at the origin and intersecting the plane
   {@code z = -1} at the corners {@code (-1, -1, -1)} and
   {@code (+1, +1, -1)}.
<p>
   The standard normalized orthographic view volume is the infinitely
   long parallelepiped centered on the z-axis and intersecting the
   xy-plane at the corners {@code (-1, -1, 0)} and {@code (+1, +1, 0)}.
<p>
   The user defined view volume (determined by the {@link Scene}'s
   {@link Camera} object) is either the infinitely long pyramid with its
   apex at the origin and intersecting the plane {@code z = -near} at the
   corners {@code (left, bottom, -near)} and {@code (right, top, -near)},
   or it is the infinitely long parallelepiped parallel to the z-axis and
   intersecting the xy-plane at the corners {@code (left, bottom, 0)}
   and {@code (right, top, 0)}.
<p>
   The view coordinate system is relative to the user defined view volume.
<p>
   The normalized camera coordinate system is relative to the normalized
   view volume.
<p>
   The transformation formulas that transform the user defined view volume
   into the normalized view volume also transforms the view coordinate
   system into the normalized camera coordinate system.
*/
public final class View2Camera
{
   /**
      Use the {@link Camera}'s view volume data to transform each
      {@link Vertex} from the {@link Camera}'s view coordinate system
      to the normalized camera coordinate system.

      @param model  {@link Model} with {@link Vertex} objects in the camera's view coordinate system
      @param camera  the {@link Scene}'s {@link Camera} with the view volume data
      @return a new {@link Model} with {@link Vertex} objects in the normalized camera coordinate system
   */
   public static Model view2camera(final Model model, final Camera camera)
   {
      final double l = camera.left; // This data defines the camera's view volume.
      final double r = camera.right;
      final double b = camera.bottom;
      final double t = camera.top;
      final double near = - camera.n;

      // A new vertex list to hold the transformed vertices.
      final List<Vertex> newVertexList = new ArrayList<>(model.vertexList.size());

      // Replace each Vertex object with one that contains
      // normalized camera coordinates.
      for (final Vertex v : model.vertexList)
      {
         double v_x, v_y, v_z;

         if ( camera.perspective )
         {
            // We use two steps to transform the camera's perspective
            // view volume into the standard perspective view volume.
            // The first step skews the current view volume so that
            // its center line is the negative z-axis (this takes an
            // asymmetric view volume and makes it symmetric). The second
            // step scales the skewed view volume so that it intersects
            // the plane z = -1 with corners (-1, -1, -1) and (+1, +1, -1)
            // (this gives the symmetric view volume a 90 degree field-of-view).

            // For each vertex, skew its coordinates so that the
            // negative z-axis is at the center of the view volume.
            v_z = v.z;
            v_x = v.x - v_z * (r + l)/(2 * near);
            v_y = v.y - v_z * (t + b)/(2 * near);

            // For each vertex, scale its coordinates so that the
            // view volume has corners (-1, -1, -1) and (+1, +1, -1).
            v_x = (2 * near * v_x)/(r - l);
            v_y = (2 * near * v_y)/(t - b);
         }
         else
         {
            // We use two steps to transform the camera's orthographic
            // view volume into the standard orthographic view volume.
            // The first step translates the current view volume so that
            // its center line is the z-axis. The second step scales the
            // translated view volume so that it intersects the plane
            // z = 0 with corners (-1, -1, 0) and (+1, +1, 0).

            // For each vertex, translate its coordinates so that
            // the z-axis is at the center of the view volume.
            v_z = v.z;
            v_x = v.x - (r + l)/2;
            v_y = v.y - (t + b)/2;

            // For each vertex, scale its coordinates so that the
            // view volume has corners (-1, -1, 0) and (+1, +1, 0).
            v_x = (2 * v_x)/(r - l);
            v_y = (2 * v_y)/(t - b);
         }

         newVertexList.add( new Vertex(v_x, v_y, v_z) );
      }

      return new Model(newVertexList,
                       model.lineSegmentList,
                       model.colorList,
                       model.name,
                       model.visible,
                       model.debug);
   }
}
